#!/opt/alt/python37/bin/python3 -bb
# coding:utf-8

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2019 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

# Script for create PHP data index

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import sys
import getopt
import os
import tempfile
from future.utils import iteritems
import clselect
from clselect.clselectexcept import ClSelectExcept
from clcommon.utils import mod_makedirs


def usage():
    print('Usage: ' + sys.argv[0] + ' [-d]')
    print('Options:')
    print(' -h | --help   : print this message')
    print(' -d | --delete : delete cache')
    print(' -n | --native : update only native version cache')


def _write_lines_to_file(filename, line_to_write):
    """
    Writes lines to file via temp
    :param filename: file name to write
    :param line_to_write: Data to write
    :return: None
    """
    if not os.path.isdir(clselect.ClSelect.CACHEFILE_DIR):
        mod_makedirs(clselect.ClSelect.CACHEFILE_DIR, 0o755)
    _, temp_file_name = tempfile.mkstemp(dir=clselect.ClSelect.CACHEFILE_DIR)
    f = open(temp_file_name, 'w')
    f.write(line_to_write)
    f.close()
    os.rename(temp_file_name, filename)
    os.chmod(filename, 0o644)


def _get_all_php_paths():
    """
    Retrives php binary path for all available php version
    :return: dictionary with php versions and paths. Example:
        {'5.1': '/opt/alt/php51/usr/bin/php',
         '5.2': '/opt/alt/php52/usr/bin/php',
         '5.3': '/opt/alt/php53/usr/bin/php',
         '5.4': '/opt/alt/php54/usr/bin/php',
         '5.5': '/opt/alt/php55/usr/bin/php',
         '5.6': '/opt/alt/php56/usr/bin/php',
         '4.4': '/opt/alt/php44/usr/bin/php',
         '7.0': '/opt/alt/php70/usr/bin/php',
         '7.1': '/opt/alt/php71/usr/bin/php',
         'native': '/usr/bin/php'}
    """
    cl_select = clselect.ClSelect()
    # Reads /etc/cl.selector/selector.conf
    return cl_select.get_all_php_binaries_paths()


def build_dat_files(only_native=False):
    """
    Generates cache files (.dat) for alt-php versions
    :return: None
    """
    try:
        all_php_paths = _get_all_php_paths()
        if only_native:
            all_php_paths = {'native': all_php_paths['native']}\
                if all_php_paths.get('native') else {}
        # cache built-in modules
        for version, php_bin_path in iteritems(all_php_paths):
            # if php binary is not exist - pass this version
            if not os.path.exists(php_bin_path):
                continue
            # Build arguments
            php_args_base = [php_bin_path,
                             '-d', 'opcache.enable_cli=0', '-d', 'display_startup_errors=0',
                             '-d', 'display_errors=1', '-d', 'error_reporting=E_ALL']
            php_args = php_args_base[:]
            php_args.append('-qm')
            if version != 'native':
                # Set PHP_INI_SCAN_DIR equal to the path to an empty directory
                # (directory absence in case if CageFS not installed suits too, doesn't generate errors)
                # to avoid loading any additional .ini files (e.g. /opt/alt/phpXX/etc/php.d/default.ini)
                env_data = {'PHP_INI_SCAN_DIR': '/usr/share/cagefs/.cagefs.empty'}
            else:
                env_data = None
            # Run php binary
            php_output = clselect.utils.run_command(php_args, env_data=env_data)
            filename_to_write = clselect.ClSelect.CACHEFILE_PATTERN % version
            _write_lines_to_file(filename_to_write, php_output)
            if version == 'native':
                php_args = php_args_base[:]
                php_args.append('-n')
                php_args.append('-v')
                php_output = clselect.utils.run_command(php_args)
                filename_to_write = clselect.ClSelect.CACHEFILE_NATIVE_VER_PATTERN
                _write_lines_to_file(filename_to_write, php_output)
    except ClSelectExcept.ConfigNotFound:
        # /etc/cl.selector/selector.conf absent - no alt-php installed - ignore
        pass
    except Exception as e:
        sys.stderr.write(str(e))


def main():
    only_native = False
    try:
        opts, _ = getopt.getopt(sys.argv[1:], 'dhn', ['delete', 'help', 'native'])
    except getopt.GetoptError as err:
        print(str(err))
        usage()
        sys.exit(2)
    try:
        for o, _ in opts:
            if o in ('-h', '--help'):
                usage()
                sys.exit(2)
            elif o in ('-d', '--delete'):
                cl_select = clselect.ClSelect()
                all_php_paths = cl_select.get_all_php_binaries_paths()
                for version in all_php_paths.keys():
                    file_name = clselect.ClSelect.CACHEFILE_PATTERN % version
                    try:
                        os.unlink(file_name)
                    except (OSError, IOError):
                        pass
                try:
                    os.unlink(clselect.ClSelect.CACHEFILE_NATIVE_VER_PATTERN)
                except (OSError, IOError):
                    pass
                sys.exit(0)
            elif o in ('-n', '--native'):
                only_native = True
            else:
                usage()
                sys.exit(2)
    except ClSelectExcept.ConfigNotFound:
        # /etc/cl.selector/selector.conf absent - no alt-php installed - ignore
        pass
    except Exception as e:
        sys.stderr.write(str(e))
    # Creates .dat file for each php version
    build_dat_files(only_native)
    sys.exit(0)

if __name__ == '__main__':
    main()
